import java.io.*;
import java.util.*;

import org.jdom2.*;
import org.jdom2.filter.Filters;
import org.jdom2.input.SAXBuilder;
import org.jdom2.output.XMLOutputter;
import org.jdom2.xpath.*;



/**
 * Takes a featuremodel.xml file generated using FeatureModelConfigurator and a given 
 * ROS/ Gazebo .gazebo.xacro file.
 * The .gazebo.xacro file is modified to support the features specified in the
 * featuremodel.xml file.
 */
public class Adapter {
	
	/**
	 * Reads an .xml file from file path and returns a JDOM document.
	 * 
	 * @param src			The .xml file path
	 * @return JDOM document
	 * @throws JDOMException
	 * @throws IOException
	 */
	public static Document readXml(String src) throws JDOMException, IOException {
		File file = new File(src);
		SAXBuilder saxBuilder = new SAXBuilder();
		Document doc = saxBuilder.build(file);

		return doc;
	}
	
	/**
	 * Executes given XPath expression for elements.
	 * 
	 * @param doc			The JDOM Document
	 * @param xPath			XPath expression
	 * @return List with all elements found or empty list if no match found
	 */
	public static List<Element> xQueryElements(Document doc, String xPath) {
		XPathFactory xFactory = XPathFactory.instance();
		XPathExpression<Element> expr = xFactory.compile(xPath, Filters.element());
		List<Element> result = expr.evaluate(doc);
		
		return result;
	}
	
	/**
	 * Executes given XPath expression for attributes.
	 * 
	 * @param doc			The JDOM Document
	 * @param xPath			XPath expression
	 * @return List with all attributes found or empty list if no match found
	 */
	public static List<Attribute> xQueryAttributes(Document doc, String xPath) {
		XPathFactory xFactory = XPathFactory.instance();
		XPathExpression<Attribute> expr = xFactory.compile(xPath, Filters.attribute());
		List<Attribute> result = expr.evaluate(doc);
		
		return result;
	}
	
	/**
	 * Searches a document for the first occurrence of an xml tag.
	 * 
	 * @param doc			The JDOM Document
	 * @param elementName	Name of the xml tag
	 * @return First element found or NULL if no match found
	 */
	public static Element searchFirstElement(Document doc, String elementName) {
		//XPath index starts with 1; [] has a higher precedence than //
		List<Element> query= xQueryElements(doc, "(//"+elementName+")[1]");
		
		if(query.isEmpty()) return null;
		return query.get(0);
	}
	
	/**
	 * Searches a document for all occurrences of an xml tag.
	 * 
	 * @param doc			The JDOM Document
	 * @param elementName	Name of the xml tag
	 * @return List with all elements found or empty list if no match found
	 */
	public static List<Element> searchAllElements(Document doc, String elementName) {
		return xQueryElements(doc, "//"+elementName);
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			//TODO##### Input ###############################
			Document featureModel = readXml("resources/input/featuremodel.xml");
			Document sensorConfig = readXml("resources/input/sensorconfig.xml");
			Document gazeboModel = null;
			
			//CustomBot not supported for now
			List<Attribute> botList = xQueryAttributes(featureModel, "//module[@types='TurtleBot']/@name");
			String botName = "";
			String fileName = "";
			
			switch(botList.size()) {
				case 0:
					System.out.println("Invalid input: No bot selected");
					System.exit(-1);
					break;
				case 1:
					botName = botList.get(0).getValue();
					break;
				default:
					System.out.println("Invalid input: Multiple bots selected");
					System.exit(-1);
			}
			switch(botName) {
				case "Burger":
					fileName = "turtlebot3_burger.gazebo.xacro";
					break;
				case "Waffle":
					fileName = "turtlebot3_waffle.gazebo.xacro";
					break;
				case "WafflePi":
					fileName = "turtlebot3_waffle_pi.gazebo.xacro";
					break;
				default:
					System.out.println("Cannot handle \'"+botName+"\': No input file known.");
					System.exit(-1);
			}
			gazeboModel = readXml("resources/input/turtlebot3-master/turtlebot3_description/urdf/"+fileName);
			//###############################################
			
			
			//TODO##### Operate on .gazebo.xacro ############
			
			//hard-coded and only for demo purposes
			List<Element> cSensorList = xQueryElements(sensorConfig, "//sensor[@type='ray' and @name='lds_lfcd_sensor']");
			Element cSensor = cSensorList.get(0);
			
			List<Element> gSensorList = xQueryElements(gazeboModel, "//sensor[@type='ray' and @name='lds_lfcd_sensor']");
			Element gSensor = gSensorList.get(0);

			String new_min = cSensor.getChild("range").getChild("min").getText();
			String new_max = cSensor.getChild("range").getChild("max").getText();
			
			gSensor.getChild("ray").getChild("range").getChild("min").setText(new_min);
			gSensor.getChild("ray").getChild("range").getChild("max").setText(new_max);
			//###############################################
			
			
			//######### Output result #######################
			FileOutputStream dest = new FileOutputStream("resources/output/"+fileName);
			XMLOutputter xmlOutput = new XMLOutputter();
			
			xmlOutput.output(gazeboModel, dest);
			//###############################################
			
			
			//XPathExamples
			XPathFactory xFactory = XPathFactory.instance();
			XPathExpression<Element> xExpr = xFactory.compile("robot/gazebo/sensor/ray/range/min", Filters.element());
			Element test = xExpr.evaluateFirst(gazeboModel);
			List<Element> test2 = searchAllElements(gazeboModel, "min");
			List<Element> test3 = searchAllElements(gazeboModel, "max");
			Element test4 = searchFirstElement(gazeboModel, "plugin");
			List<Attribute> test5 = xQueryAttributes(gazeboModel, "(//plugin)[1]/@*");
			
			if(test != null) xmlOutput.output(test, System.out);
			System.out.println("\n");
			xmlOutput.output(test2, System.out);
			System.out.println("\n");
			xmlOutput.output(test3, System.out);
			System.out.println("\n");
			if(test4 != null) xmlOutput.output(test4, System.out);
			System.out.println("\n");
			for(Attribute a : test5) System.out.println(a.getName()+"="+a.getValue());
			
		} catch (JDOMException | IOException e) {
			e.printStackTrace();
		}
	}
}
